#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <string>
#include <sys/stat.h>
#include <unistd.h>
#include <limits>

#include "thaecS.h"
#include "curlget.h"
#include "SpiceUsr.h"
#include "SpiceZfc.h"
#include "SpiceSPK.h"

int timeLocationLoop(TimeInstant ti, const TerrestrialLocation &tl);

void time_dv::toSofaJd(sofa_jd& jd) const 
{ int error=iauDtf2d("UT0", yr, mo, dy, hr, mi, sc, &jd[0], &jd[1]); if(error)exept(error); }
void sofa_jd::toTime(time_dv& dv, int prec=6) const
{ int error=iauD2dtf("UT0", prec, jd[0], jd[1], &dv.yr, &dv.mo, &dv.dy, dv.hmsf); if(error)exept(error); }

TimeInstant FindUpperCulmination(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps=60);
TimeInstant FindLowerCulmination(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps=60);

TimeInstant FindRise(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps=60);
TimeInstant FindSet(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps=60);

void CelestalPosition(GenPosVelVec pbpv, iauASTROM astrom, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob);
void GetTerrestrialObserverVectors(const TimeInstant &ti, const TerrestrialLocation &tl, iauASTROM &astrom, GenPosVelVec &obpv, GenPosVelVec &ebpv);
void GetTargetVectors(const char *target, const double spiceT, GenPosVelVec obpv, GenPosVelVec ebpv, GenPosVelVec &pbpv, GenPosVelVec &pepv);
void DownloadFiles();
void DownloadEphemerides();
double ReadTTcorr(const double mjd_utc);
double ReadDAT(const double jd_utc, const double mjd_utc);
double ReadDUT(const double mjd, double &xp, double &yp, double &dX, double &dY);
double ReadDT(const double mjd, const double year);
double CalculateDT(const double year);

// run with current local time:  ./thaec
// run with some given UTC time: ./thaec 21.10.2014 18:23:50
// run with some given UTC time: ./thaec 2015-10-21 18:23:50
int main(int argc, char* argv[])
{
  DownloadFiles();printf("\n");
  DownloadEphemerides();printf("\n");

  // point towards the SPICE files, multiple calls are allowed
  // at files with overlaps the last file loaded has the priority
  furnsh_c ( "data/de440s.bsp" ); 
  furnsh_c ( "data/pck000xx.tpc" );
  furnsh_c ( "data/earth_fixed.tf" );
  // furnsh_c ( "data/mysites.bsp" );
  // furnsh_c ( "data/mysites.tf" );
  // furnsh_c ( "data/fov.ti" );
  // furnsh_c ( "data/earth_latest_high_prec.bpc" );

  // your location on earth (east is +), elevation (m), temperature (deg C), pressure (hPa), and humidity (%), wavelength (micrometer, 0.38 to 0.78 for visible light, 0.78-100 for IR and exactly 101 for radio)
  TerrestrialLocation tl;
  tl.lat=50.11552*DPI/180; tl.lon=8.68417*DPI/180; tl.elev=113; tl.temp=20; tl.press=1013; tl.relhum=0.7; tl.wavelength=0.555;

  time_t rawtime;
  struct tm *ptm;
 
  time(&rawtime);
  //ptm = localtime(&rawtime);
  //printf("local:      %02i.%02i.%i %02i:%02i:%02i\n", ptm->tm_mday, ptm->tm_mon+1, ptm->tm_year+1900, ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
  ptm = gmtime(&rawtime);
    
  for(int i = 1; i<argc; i++){
    std::string str(argv[i]);
    size_t plz;

    plz = str.find_first_of(".");
    if(plz!=std::string::npos){
      ptm->tm_mday = atoi(str.substr(plz-2,2).c_str());
      ptm->tm_mon = atoi(str.substr(plz+1,2).c_str())-1;
      ptm->tm_year = atoi(str.substr(plz+4,4).c_str())-1900;
      ptm->tm_hour = 0;
      ptm->tm_min = 0;
      ptm->tm_sec = 0;
    }

    plz = str.find_first_of("-");
    if(plz!=std::string::npos){
      ptm->tm_year = atoi(str.substr(plz-4,4).c_str())-1900;
      ptm->tm_mon = atoi(str.substr(plz+1,2).c_str())-1;
      ptm->tm_mday = atoi(str.substr(plz+4,4).c_str());
      ptm->tm_hour = 0;
      ptm->tm_min = 0;
      ptm->tm_sec = 0;
    }

    plz = str.find_first_of(":");
    if(plz!=std::string::npos){
      ptm->tm_hour = atoi(str.substr(plz-2,2).c_str());
      ptm->tm_min = atoi(str.substr(plz+1,2).c_str());
      ptm->tm_sec = 0;
    }
    plz = str.find_first_of(":",plz+1);
    if(plz!=std::string::npos)
      ptm->tm_sec = atoi(str.substr(plz+1,2).c_str());
  }
  rawtime = timegm(ptm);
  
  while(true){
    printf("Unix time:  %ld\n",rawtime);
    ptm = gmtime(&rawtime);

    TimeInstant ti;
    time_dv dv;

    dv.yr = ptm->tm_year+1900;
    dv.mo = ptm->tm_mon+1;
    dv.dy = ptm->tm_mday;
    dv.hr = ptm->tm_hour;
    dv.mi = ptm->tm_min;
    dv.sc = ptm->tm_sec;
    dv.fr = 0;

    dv.toSofaJd(ti.utc);
    int cont = timeLocationLoop(ti,tl);

    switch(cont){
    case kPrevSec: rawtime--; break;
    case kNextSec: rawtime++; break;
    case kPrevMin: rawtime-=60; break;
    case kNextMin: rawtime+=60; break;
    case kPrevHour: rawtime-=3600; break;
    case kNextHour: rawtime+=3600; break;
    case kPrevDay: rawtime-=3600*24; break;
    case kNextDay: rawtime+=3600*24; break;
    case kPrevWeek: rawtime-=3600*24*7; break;
    case kNextWeek: rawtime+=3600*24*7; break;
    case kPrevYear: ptm->tm_year--; rawtime=timegm(ptm); break;
    case kNextYear: ptm->tm_year++; rawtime=timegm(ptm); break;
    }
  }

  return 0;
}

int timeLocationLoop(TimeInstant ti, const TerrestrialLocation &tl){
  time_dv dv;
  double ut;

  ti.utc.toTime(dv);
  double mjd_utc = ti.utc.MJD();
  double jd_utc  = ti.utc.JD();
  ti.dut = ReadDUT(mjd_utc,ti.xp,ti.yp,ti.dX,ti.dY);
  ti.dat = ReadDAT(jd_utc, mjd_utc);
  ti.ctt = ReadTTcorr(mjd_utc);

  if(ti.dut!=0 && ti.dat!=0){
    printf("UTC:        %02i.%02i.%i %02i:%02i:%02i\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc);

    ti.utc2ut1();
    ti.ut1.toTime(dv);
    printf("UT1:        %02i.%02i.%04i %02i:%02i:%02i.%06i  (UT1-UTC: %.6fs)\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr, ti.dut);

    ti.utc2tai();
    ti.tai.toTime(dv);
    printf("TAI:        %02i.%02i.%04i %02i:%02i:%02i.%06i  (TAI-UTC: %.6fs)\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr, ti.dat);
    
    ti.tai2tt();
    ti.tt.toTime(dv);
    ti.dt = (ti.tt-ti.ut1)*24*3600;
  }else{
    printf("UT1:        %02i.%02i.%04i %02i:%02i:%02i\n",  dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc);

    ti.ut1=ti.utc;
    ti.tai=ti.utc;
    ti.dut=ti.dat=ti.xp=ti.yp=ti.dX=ti.dY=0;
    ti.dt = ReadDT(mjd_utc, dv.yr+(dv.mo-1.0)/12+(dv.dy-1.0)/30/12);

    ti.ut2tt();
    ti.tt.toTime(dv);
  }
  printf("TT(=TDT):   %02i.%02i.%04i %02i:%02i:%02i.%06i  (TT-UT1:  %.6fs)\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr, ti.dt);
  ti.tt.toTime(dv);
  ti.dt = (ti.tt-ti.ut1)*24*3600;
  printf("TT(BIPMxx): %02i.%02i.%04i %02i:%02i:%02i.%06i  (TTcorr:  %.4fus)\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr, ti.ctt);

  ti.tt2tcg();
  ti.tcg.toTime(dv);
  printf("TCG:        %02i.%02i.%04i %02i:%02i:%02i.%06i\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr);

  ut = ti.ut1.dayFraction();
  ti.dtr = iauDtdb(ti.tt[0], ti.tt[1], ut, 0, 0, 0);
  ti.tt2tdb();
  ti.dtr = iauDtdb(ti.tdb[0], ti.tdb[1], ut, 0, 0, 0);  //iterative method
  ti.tt2tdb();
  ti.tdb.toTime(dv);
  printf("TDB:        %02i.%02i.%04i %02i:%02i:%02i.%06i  (TDB-TT:  %6.6fms)\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr, ti.dtr*1000);
  ti.tdb2spice();

  double pv[6], lighttime;
  char defact[80], defprt[80];
  erract_c ( "GET", 80, defact );
  errprt_c ( "GET", 80, defprt );
  erract_c ( "SET", 0, "RETURN" );
  errprt_c ( "SET", 0, "NONE" );
  //try to read TT-TDB from the file
  spkezr_c ( "1000000001", ti.spiceT, "J2000", "NONE", "1000000000", pv, &lighttime );
  if(failed_c()){
    reset_c();
  }else{
    ti.dtr=-pv[0];
    ti.tt2tdb();
    ti.tdb.toTime(dv);
    printf("TDB(exact): %02i.%02i.%04i %02i:%02i:%02i.%06i  (TDB-TT:  %6.6fms)\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr, ti.dtr*1000);
    ti.tdb2spice();
  }
  erract_c ( "SET", 0, defact );
  errprt_c ( "SET", 0, defprt );

  ti.tdb2tcb();
  ti.tcb.toTime(dv);
  printf("TCB:        %02i.%02i.%04i %02i:%02i:%02i.%06i\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr);

  printf("SpiceT:     %14.6f\n",ti.spiceT);
  printf("JD (TT):    %14.6f\n",ti.tt.JD());
  printf("MJD (UTC):  %14.6f\n",mjd_utc);

  ti.ut12era();
  int ihmsf[4];
  ihmsf[0]=ti.era/D2PI*24;
  ihmsf[1]=fmod(ti.era/D2PI*24,1.0)*60;
  ihmsf[2]=fmod(ti.era/D2PI*24*60,1.0)*60;
  ihmsf[3]=fmod(ti.era/D2PI*24*3600,1.0)*1000000;
  printf("ERA:        %02ih %02im %02i.%06is\n", ihmsf[0], ihmsf[1], ihmsf[2], ihmsf[3]);
  
  iauXys06a(ti.tt[0], ti.tt[1], &ti.X, &ti.Y, &ti.s); //X,Y and s from 2006 precession and 2000A nutation model
  ti.sp=iauSp00(ti.tt[0], ti.tt[1]);
  printf("CIP @ GCRS: X=%e Y=%e dX=%f\" dY=%f\"\n",ti.X,ti.Y,ti.dX*3600*180/DPI,ti.dY*3600*180/DPI);
  printf("CIO @ GCRS: s=%fmas\n",ti.s*1000*3600*180/DPI);
  printf("CIP @ ITRS: x=%f\" y=%f\"\n",ti.xp*3600*180/DPI,ti.yp*3600*180/DPI);
  printf("TIO @ ITRS: s'=%fmas\n",ti.sp*1000*3600*180/DPI);
  printf("\n");
  ti.X+=ti.dX; //apply IERS correction of the 2000A nutation model
  ti.Y+=ti.dY; //apply IERS correction of the 2000A nutation model

  GenPosVelVec ebpv, obpv; //eb: earth-SSB, eh: earth-Sun, ob:observer-SSB
  iauASTROM astrom;

  GetTerrestrialObserverVectors(ti,tl,astrom, obpv, ebpv);
  obpv.convertUnits(GenPosVelVec::kAU,GenPosVelVec::kKM);
  
  // Use SPICE directly as alternative to get the observers coordinates: (take a look at data/mysites.tf)
  // spkpos_c ( "FRANKFURT", spiceT, "J2000", "NONE", "SSB", obpv.split[0], &lighttime );

  printf("Geocenter BCRS: x=%+010.7f y=%+010.7f z=%+010.7f\n",ebpv[0],ebpv[1],ebpv[2]);
  printf("Observer  BCRS: x=%+010.7f y=%+010.7f z=%+010.7f\n",astrom.eb[0],astrom.eb[1],astrom.eb[2]);

  //std::string objstr[]={"SUN","MOON","MERCURY BARYCENTER","VENUS BARYCENTER","MARS BARYCENTER","JUPITER BARYCENTER","SATURN BARYCENTER","URANUS BARYCENTER","NEPTUNE BARYCENTER","PLUTO BARYCENTER"};
  while(1){
    char objstr[80];
    int objidx, planetidx, found;
    printf("\nEnter object name or ID: ");
    scanf("%s",objstr);
    //fgets(objstr,80,stdin);
    //printf("len %i |%s|\n",strlen(objstr),objstr);
    if(strlen(objstr)==1){
      char c = objstr[0];
      if( c < '0' || c > '9')
        return c;
    }
    bods2c_c(objstr,&objidx,&found);
    if(!found){
      printf("Object not found!\n");
      printf("Try something like: Sun, Moon, Venus, Jupiter, Saturn...\n");
      continue;
    }
    bodc2s_c(objidx,80,objstr);
    if(objidx>0&&objidx<10)planetidx=100*objidx+99;
    else planetidx=objidx;
    {
      char ident[80];
      double descr[5];
      int handle;
      erract_c ( "SET", 0, "RETURN" );
      errprt_c ( "SET", 0, "NONE" );
      spksfs_c(objidx,ti.spiceT,80,&handle,descr,ident,&found);
      if(!found){
        reset_c();
        printf("Object not available...\n");
        printf("Trying to find the correspondent planetary barycenter.\n");
        if(objidx>0 && objidx<1000 && objidx%100==99){
          objidx=objidx/100;
          bodc2s_c(objidx,80,objstr);
          spksfs_c(objidx,ti.spiceT,80,&handle,descr,ident,&found);
        }
        printf("Found? %s\n",found?"yes":"no");
        if(!found) continue;
      }
      erract_c ( "SET", 0, defact );
      errprt_c ( "SET", 0, defprt );
    }
    printf("\nObject: %s (ID: %d)\n",objstr,objidx);
    
    GenPosVelVec pbpv, pepv, popv; //pb: planet-SSB, pe: planet-earth, po: planet-observer
    
    // // calculate pv of target center with respect to SSB and lighttime between geocenter and target center
    // // we do not use the aberration correction (+S) because this is included in iauAtciq
    // spkpos_c ( objstr, ti.spiceT, "J2000", "CN", "EARTH", pepv.split[0], &lighttime );
    // pepv.convertUnits(GenPosVelVec::kKM,GenPosVelVec::kAU);
    // // calculate pv of target center with respect to SSB
    // iauPpp(pepv(0),ebpv(0),pbpv(0)); //pb=pe+eb 
    
    // calculate pv of target center with respect to SSB and lighttime between observer and target center 
    // we do not use the aberration correction (+S) because this is included in iauAtciq
    spkcpo_c ( objstr, ti.spiceT, "J2000", "CENTER", "CN", obpv.split[0], "SSB", "J2000", popv.joined, &lighttime );
    // calculate pv of target center with respect to SSB
    iauPpp(popv(0),obpv(0),pbpv(0)); //pb=po+ob
    iauPmp(pbpv(1),pbpv(1),pbpv(1)); //pb=pb-pb=0 //we do not want the speed
    popv.convertUnits(GenPosVelVec::kKM,GenPosVelVec::kAU);
    pbpv.convertUnits(GenPosVelVec::kKM,GenPosVelVec::kAU);
    
    // // calculate lighttime between observer and target center
    // // we do not use the aberration correction (+S) because this is included in iauAtciq
    // spkcpo_c ( objstr, ti.spiceT, "J2000", "CENTER", "CN", obpv.split[0], "SSB", "J2000", popv.joined, &lighttime );
    // // calculate pv of target center with respect to SSB, with given lighttime correction
    // // we do not use the aberration correction (+S) because this is included in iauAtciq
    // spkpos_c ( objstr, ti.spiceT-lighttime, "J2000", "NONE", "SSB", pbpv.split[0], &lighttime );
    // pbpv.convertUnits(GenPosVelVec::kKM,GenPosVelVec::kAU);

    std::string lighttimestr="days";
    double lighttimeval=lighttime/24/3600;
    if(lighttimeval*24<24){
      lighttimeval*=24;
      lighttimestr="hours";
    }
    if(lighttimeval*60<60){
      lighttimeval*=60;
      lighttimestr="minutes";
    }
    if(lighttimeval*60<60){
      lighttimeval*=60;
      lighttimestr="seconds";
    }
    
    // // coordinates for graphics output in pixels
    // // perhaps with: http://stackoverflow.com/questions/13745093/a-portable-function-to-create-a-bmp-file-from-raw-bytes
    // iauRxpv(eclRot,pbpv.split,eclpv);
    // iauP2s(eclpv[0],&ecllon,&ecllat,&dist);
    // printf("r=%f x=%f y=%f z=%f\n",dist*10,eclpv[0][0]*10,eclpv[0][1]*10,eclpv[0][2]*10);

    Angle_t ra, dec, aob, zob;
    CelestalPosition(pbpv, astrom, ra, dec, aob, zob);
    ti.utc.toTime(dv);
    printf("Obsv:  UTC: %02i.%02i.%04i %02i:%02i:%02i ", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc);
    printf("RA: %02dh%02dm%05.2fs dec:%+03d°%02d\'%05.2f\", Az: %03d°%02d\'%05.2f\" Alt: %+03d°%02d\'%05.2f\"\n",ra.h(),ra.m(),ra.s(),dec.deg(),dec.arcM(),dec.arcS(),aob.deg(),aob.arcM(),aob.arcS(),zob.deg(),zob.arcM(),zob.arcS());
    // Use SPICE directly as alternative: (take a look at data/mysites.tf)
    // spkpos_c ( objstr, spiceT, "FRANKFURT_TOPO", "CN+S", "EARTH", pbpv.split[0], &lighttime );
    // reclat_c ( pbpv.split[0], &rv, &aob, &zob);
    // printf("obs:  Az: %09.6f Alt: %+010.6f\n",fmod(360-aob*180/DPI,360),zob*180/DPI);
    
    TimeInstant rise = FindRise(ti,tl,objstr,ra, dec, aob, zob);
    rise.utc.toTime(dv);
    printf("Rise @ UTC: %02i.%02i.%04i %02i:%02i:%02i ", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc);
    printf("RA: %02dh%02dm%05.2fs dec:%+03d°%02d\'%05.2f\", Az: %03d°%02d\'%05.2f\" Alt: %+03d°%02d\'%05.2f\"\n",ra.h(),ra.m(),ra.s(),dec.deg(),dec.arcM(),dec.arcS(),aob.deg(),aob.arcM(),aob.arcS(),zob.deg(),zob.arcM(),zob.arcS());

    TimeInstant transit = FindUpperCulmination(rise,tl,objstr,ra, dec, aob, zob);
    transit.utc.toTime(dv);
    printf("Trans@ UTC: %02i.%02i.%04i %02i:%02i:%02i ", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc);
    printf("RA: %02dh%02dm%05.2fs dec:%+03d°%02d\'%05.2f\", Az: %03d°%02d\'%05.2f\" Alt: %+03d°%02d\'%05.2f\"\n",ra.h(),ra.m(),ra.s(),dec.deg(),dec.arcM(),dec.arcS(),aob.deg(),aob.arcM(),aob.arcS(),zob.deg(),zob.arcM(),zob.arcS());

    TimeInstant set = FindSet(transit,tl,objstr,ra, dec, aob, zob);
    set.utc.toTime(dv);
    printf("Set  @ UTC: %02i.%02i.%04i %02i:%02i:%02i ", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc);
    printf("RA: %02dh%02dm%05.2fs dec:%+03d°%02d\'%05.2f\", Az: %03d°%02d\'%05.2f\" Alt: %+03d°%02d\'%05.2f\"\n",ra.h(),ra.m(),ra.s(),dec.deg(),dec.arcM(),dec.arcS(),aob.deg(),aob.arcM(),aob.arcS(),zob.deg(),zob.arcM(),zob.arcS());

    Angle_t vis=(set.tai-rise.tai)*DPI*2;
    printf("Visibility: %02ih %02im %02.0fs\n",vis.h(),vis.m(),vis.s());

    printf("Distance: %f AU (%f Gm, %f light-%s)\n", iauPm(popv(0)), iauPm(popv(0))*GenPosVelVec::kAU/1000000000, lighttimeval, lighttimestr.c_str());

    if(objidx==0)continue;
    GenPosVelVec radii;
    int dim;
    Angle_t angle;
    bodvcd_c(planetidx, "RADII", 3, &dim, radii(0));
    radii.convertUnits(GenPosVelVec::kKM,GenPosVelVec::kAU);
    double radius=(radii[0]+radii[1]+radii[2])/3;
    angle()=asin(2*radius/iauPm(popv(0))); //asin instead of atan because of: https://en.wikipedia.org/wiki/Angular_diameter
    printf("Apparent size: %d\u00B0%02d\'%05.2f\"\n",angle.deg(),angle.arcM(),angle.arcS());

    if(objidx==10)continue;
    GenPosVelVec hppv; //hppv: SUN->planet
    spkpos_c ( "SUN", ti.spiceT, "J2000", "CN+S", objstr, hppv.split[0], &lighttime );
    angle()=DPI-iauSepp(popv(0),hppv(0));
    printf("Phase angle: %3d\u00B0%02d\'%05.2f\"\n",angle.deg(),angle.arcM(),angle.arcS());
    //printf("phase angle %f\n",phaseq_c(spiceT,objstr,"SUN","EARTH","CN+S")*180/DPI);

    double loct=ti.spiceT-lighttime, pm, lambda, polrot[3][3], northvec[3]={0,0,1};
    bodeul_(&planetidx, &loct, &ra(), &dec(), &pm, &lambda);
    //printf("RA: %09.6f dec: %+010.6f\n",ra()*12/DPI,dec()*180/DPI);
    eul2m_c(-ra()-lambda,dec()-DPI/2,ra(),3,2,3,polrot);
    iauRxp(polrot,northvec,northvec); //alt: radrec_c(1,ra,dec,northvec)
    // iauC2s(northvec,&ra,&dec);
    // printf("RA: %09.6f dec: %+010.6f\n",ra()*12/DPI,dec()*180/DPI);
    angle()=iauSepp(popv(0),northvec)-DPI/2;
    printf("Equator viewing angle: %+d\u00B0%02d\'%05.2f\"\n",angle.deg(),angle.arcM(),angle.arcS());
  }
  return 0;
}

void CelestalPosition(GenPosVelVec pbpv, iauASTROM astrom, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob){
  double pmr, pmd, px, rv;
  Angle_t hob;
  iauPvstar(pbpv.split,&ra(),&dec(),&pmr,&pmd,&px,&rv);
  //printf("BCRS: RA: %02dh%02dm%05.2fs dec:%+03d°%02d\'%05.2f\", x=%+010.7f y=%+010.7f z=%+010.7f\n",ra.h(),ra.m(),ra.s(),dec.deg(),dec.arcM(),dec.arcS(),pbpv[0],pbpv[1],pbpv[2]);
  iauAtciq(ra(),dec(),0,0,px,0,&astrom,&ra(),&dec());
  //printf("CIRS: RA: %02dh%02dm%05.2fs dec:%+03d°%02d\'%05.2f\"\n",ra.h(),ra.m(),ra.s(),dec.deg(),dec.arcM(),dec.arcS());
  iauAtioq(ra(),dec(),&astrom,&aob(),&zob(),&hob(),&dec(),&ra()); zob()=DPI/2-zob();
  //printf("obsv: RA: %02dh%02dm%05.2fs dec:%+03d°%02d\'%05.2f\", Az: %03d°%02d\'%05.2f\" Alt: %+03d°%02d\'%05.2f\"\n",ra.h(),ra.m(),ra.s(),dec.deg(),dec.arcM(),dec.arcS(),aob.deg(),aob.arcM(),aob.arcS(),zob.deg(),zob.arcM(),zob.arcS());
}

void GetTerrestrialObserverVectors(const TimeInstant &ti, const TerrestrialLocation &tl, iauASTROM &astrom, GenPosVelVec &obpv, GenPosVelVec &ebpv){
  double lighttime;
  GenPosVelVec ehpv;

  // Barycentric coordinates and velocities of the earth
  spkezr_c ( "EARTH", ti.spiceT, "J2000", "NONE", "SSB", ebpv.joined, &lighttime );
  ebpv.convertUnits(GenPosVelVec::kKM,GenPosVelVec::kAU);

  // Heliocentric coordinates of the earth
  spkezr_c ( "EARTH", ti.spiceT, "J2000", "NONE", "SUN", ehpv.joined, &lighttime );
  ehpv.convertUnits(GenPosVelVec::kKM,GenPosVelVec::kAU);

  iauRefco(tl.press,tl.temp,tl.relhum,tl.wavelength,&astrom.refa,&astrom.refb);
  iauApco(ti.tdb[0],ti.tdb[1],ebpv.split,ehpv.split[0],ti.X,ti.Y,ti.s,ti.era,tl.lon,tl.lat,tl.elev,ti.xp,ti.yp,ti.sp,astrom.refa,astrom.refb,&astrom);
  for(int i2=0; i2<3; i2++)
    obpv(0)[i2]=astrom.eb[i2];
  for(int i2=0; i2<3; i2++)
    obpv(1)[i2]=astrom.v[i2];
  obpv.convertUnitsVel(GenPosVelVec::kLS,GenPosVelVec::kAU);
}

void GetTargetVectors(const char *target, const double spiceT, GenPosVelVec obpv, GenPosVelVec ebpv, GenPosVelVec &pbpv, GenPosVelVec &pepv){
  double lighttime;

  // // calculate pv of target center with respect to SSB and lighttime between geocenter and target center
  // // we do not use the aberration correction (+S) because this is included in iauAtciq
  spkpos_c ( target, spiceT, "J2000", "CN", "EARTH", pepv.split[0], &lighttime );
  pepv.convertUnits(GenPosVelVec::kKM,GenPosVelVec::kAU);
  iauPpp(pepv(0),ebpv(0),pbpv(0)); //pb=pe+eb   //we need SSB coordinates of the target
}

double AboveGeometricHorizon(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target){
  GenPosVelVec obpv, ebpv, pbpv, pepv, oepv; //ob: obs-SSB, eb: earth-SSB, pb: planet-SSB, pe: planet-earth, oe: obs-earth
  iauASTROM astrom;
  
  GetTerrestrialObserverVectors(ti,tl,astrom, obpv, ebpv);
  GetTargetVectors(target, ti.spiceT, obpv, ebpv, pbpv, pepv);
  iauPmp(obpv(0),ebpv(0),oepv(0)); //oe=ob-eb
  return iauPdp(oepv(0),pepv(0));
}

double AboveHorizon(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob){
  GenPosVelVec obpv, ebpv, pbpv, pepv, oepv; //ob: obs-SSB, eb: earth-SSB, pb: planet-SSB, pe: planet-earth, oe: obs-earth
  iauASTROM astrom;

  GetTerrestrialObserverVectors(ti,tl,astrom, obpv, ebpv);
  GetTargetVectors(target, ti.spiceT, obpv, ebpv, pbpv, pepv);
  
  CelestalPosition(pbpv, astrom, ra, dec, aob, zob);
  return zob();
}

TimeInstant FindUpperLowerCulmination(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps, int upLow){
  TimeInstant rise = ti;
  GenPosVelVec obpv, ebpv, pbpv, pepv, oepv; //ob: obs-SSB, eb: earth-SSB, pb: planet-SSB, pe: planet-earth, oe: obs-earth

  while(eps>1){
    double alt;
    double prev = -1;
    while((alt=AboveGeometricHorizon(rise,tl,target))>prev){
      // printf("+++ %f min\n",eps);
      // printf("%e %e\n",alt,prev);
      // time_dv dv;
      // rise.tdb.toTime(dv);
      // printf("   :        %02i.%02i.%04i %02i:%02i:%02i.%06i\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr);
      prev = alt;
      rise.tdb+=upLow*eps/60/24;
      rise.era+=upLow*eps/60/24*D2PI*1.00273781191135448;
      rise.tdb2spice();
    }
    // printf("--- %f min\n",eps);
    // printf("%e %e\n",alt,prev);
    // time_dv dv;
    // rise.tdb.toTime(dv);
    // printf("   :        %02i.%02i.%04i %02i:%02i:%02i.%06i\n", dv.dy, dv.mo, dv.yr, dv.hr, dv.mi, dv.sc, dv.fr);
    eps/=2;
    rise.tdb-=upLow*3*eps/60/24;
    rise.era-=upLow*3*eps/60/24*D2PI*1.00273781191135448;
    rise.tdb2spice();
  } 
  rise.tdb2spice();
  rise.tdb2tcb();
  rise.tdb2tt();
  rise.tt2tai();
  rise.tai2utc();
  rise.utc2ut1();
  rise.ut12era();

  while(eps>0.001){
    double alt;
    double prev = -10;
    while((alt=AboveHorizon(rise,tl,target,ra,dec,aob,zob))>prev){
      prev = alt;
      rise.tdb+=upLow*eps/60/24;
      rise.era+=upLow*eps/60/24*D2PI*1.00273781191135448;
      rise.tdb2spice();
    }
    eps/=2;
    rise.tdb-=upLow*3*eps/60/24;
    rise.era-=upLow*3*eps/60/24*D2PI*1.00273781191135448;
    rise.tdb2spice();
  } 
  rise.tdb2spice();
  rise.tdb2tcb();
  rise.tdb2tt();
  rise.tt2tai();
  rise.tai2utc();
  rise.utc2ut1();
  rise.ut12era();

  return rise;
}
TimeInstant FindUpperCulmination(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps){
  return FindUpperLowerCulmination(ti, tl, target, ra, dec, aob, zob, eps, 1);
}
TimeInstant FindLowerCulmination(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps){
  // returns wrong results
  return FindUpperLowerCulmination(ti, tl, target, ra, dec, aob, zob, eps, -1);
}

TimeInstant FindRiseSet(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps, int riseSet){
  TimeInstant rise = ti;
  GenPosVelVec obpv, ebpv, pbpv, pepv, oepv; //ob: obs-SSB, eb: earth-SSB, pb: planet-SSB, pe: planet-earth, oe: obs-earth
  while(eps>1){
    while(AboveGeometricHorizon(rise,tl,target)>0){
      rise.tdb+=riseSet*eps/60/24;
      rise.era+=riseSet*eps/60/24*D2PI*1.00273781191135448;
      rise.tdb2spice();
    }
    while(AboveGeometricHorizon(rise,tl,target)<0){
      rise.tdb-=riseSet*eps/60/24;
      rise.era-=riseSet*eps/60/24*D2PI*1.00273781191135448;
      rise.tdb2spice();
    }
    eps/=2;
  }
  rise.tdb2tcb();
  rise.tdb2tt();
  rise.tt2tai();
  rise.tai2utc();
  rise.utc2ut1();
  rise.ut12era();

  while(eps>0.001){
    while(AboveHorizon(rise,tl,target,ra,dec,aob,zob)>0){
      rise.tdb+=riseSet*eps/60/24;
      rise.era+=riseSet*eps/60/24*D2PI*1.00273781191135448;
      rise.tdb2spice();
    }
    while(AboveHorizon(rise,tl,target,ra,dec,aob,zob)<0){
      rise.tdb-=riseSet*eps/60/24;
      rise.era-=riseSet*eps/60/24*D2PI*1.00273781191135448;
      rise.tdb2spice();
    }
    eps/=2;
  }
  rise.tdb2tcb();
  rise.tdb2tt();
  rise.tt2tai();
  rise.tai2utc();
  rise.utc2ut1();
  rise.ut12era();

  return rise;
}
TimeInstant FindRise(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps){
  return FindRiseSet(ti, tl, target, ra, dec, aob, zob, eps, -1);
}
TimeInstant FindSet(const TimeInstant &ti, const TerrestrialLocation &tl, const char *target, Angle_t &ra, Angle_t &dec, Angle_t &aob, Angle_t &zob, double eps){
  return FindRiseSet(ti, tl, target, ra, dec, aob, zob, eps, 1);
}

double ReadTTcorr(const double mjd_utc){
  if(mjd_utc<43144) return 0; //before 01.01.1977 TDT/TT did not exist
  int imjd = mjd_utc, tmjd, fmjd=0, smjd=0;
  double dut1 = 0, dut2 = 0;

  std::string line;
  std::ifstream myfile2("corrections/TTBIPM.act");
  if(myfile2.is_open())
    {
      while(myfile2.good())
        {
          getline(myfile2,line);
          if(line.length()>24 && line.length()<27){
            tmjd = atoi(line.substr(0,5).c_str());
            if(tmjd <= imjd){
              dut1 = atof(line.substr(17,-1).c_str());
              fmjd = tmjd;
              continue;
            }
            if(tmjd > imjd){
              dut2 = atof(line.substr(17,-1).c_str());
              smjd = tmjd;
              break;
            }
          }
        }
      myfile2.close();
    }
  if(smjd==0) return dut1;
  return (mjd_utc-fmjd)/(smjd-fmjd)*dut2+(smjd-mjd_utc)/(smjd-fmjd)*dut1;
}

double ReadDAT(const double jd_utc, const double mjd_utc){
  double ijd = jd_utc, tjd, fmjd=0;
  double dat = 0, drift = 0;

  std::string line;
  std::ifstream myfile2("corrections/tai-utc.dat");
  if(myfile2.is_open())
    {
      while(myfile2.good())
        {
          getline(myfile2,line);
          if(line.length()>8){
            tjd = atof(line.substr(17,9).c_str());
            if(tjd <= ijd){
              dat = atof(line.substr(37,11).c_str());
              fmjd= atof(line.substr(60,5).c_str());
              drift=atof(line.substr(70,9).c_str());
              continue;
            }
          }
        }
      myfile2.close();
    }
  return dat+(mjd_utc-fmjd)*drift;
}

double ReadDUT(const double mjd_utc, double &xp, double &yp, double &dX, double &dY){
  int imjd = mjd_utc, tmjd, fmjd=0, smjd=0;
  double dut1 = 0, dut2 = 0, x1, x2, y1, y2, X1, X2, Y1, Y2;
  std::string line;

  std::ifstream myfile2("corrections/eopc04_IAU2000.62-now");
  if(myfile2.is_open())
    {
      while(myfile2.good())
        {
          getline(myfile2,line);
          if(line.length()>80){
            tmjd = atoi(line.substr(14,5).c_str());
            if(tmjd == imjd && dut1==0){
              x1 = atof(line.substr(21,9).c_str());
              y1 = atof(line.substr(32,9).c_str());
              X1 = atof(line.substr(67,9).c_str());
              Y1 = atof(line.substr(78,9).c_str());
              dut1 = atof(line.substr(43,10).c_str());
              fmjd = tmjd;
              continue;
            }
            if(tmjd == imjd+1){
              x2 = atof(line.substr(21,9).c_str());
              y2 = atof(line.substr(32,9).c_str());
              X2 = atof(line.substr(67,9).c_str());
              Y2 = atof(line.substr(78,9).c_str());
              dut2 = atof(line.substr(43,10).c_str());
              smjd = tmjd;
              break;
            }
          }
        }
      myfile2.close();
    }

  std::ifstream myfile3("corrections/finals2000A.daily");
  if(dut2==0 && myfile3.is_open())
    {
      while(myfile3.good())
        {
          getline(myfile3,line);
          if(line.length()>80){
            tmjd = atoi(line.substr(7,5).c_str());
            if(tmjd == imjd && dut1==0){
              x1 = atof(line.substr(18,9).c_str());
              y1 = atof(line.substr(37,9).c_str());
              X1 = atof(line.substr(97,9).c_str())/1000;
              Y1 = atof(line.substr(116,9).c_str())/1000;
              dut1 = atof(line.substr(58,10).c_str());
              fmjd = tmjd;
              continue;
            }
            if(tmjd == imjd+1){
              x2 = atof(line.substr(18,9).c_str());
              y2 = atof(line.substr(37,9).c_str());
              X2 = atof(line.substr(97,9).c_str())/1000;
              Y2 = atof(line.substr(116,9).c_str())/1000;
              dut2 = atof(line.substr(58,10).c_str());
              smjd = tmjd;
              break;
            }
          }
        }
      myfile3.close();
    }

  std::ifstream myfile("corrections/finals2000A.data");
  if(dut2==0 && myfile.is_open())
    {
      while(myfile.good())
        {
          getline(myfile,line);
          if(line.length()>80){
            tmjd = atoi(line.substr(7,5).c_str());
            if(tmjd == imjd && dut1==0){
              x1 = atof(line.substr(18,9).c_str());
              y1 = atof(line.substr(37,9).c_str());
              X1 = atof(line.substr(97,9).c_str())/1000;
              Y1 = atof(line.substr(116,9).c_str())/1000;
              dut1 = atof(line.substr(58,10).c_str());
              fmjd = tmjd;
              continue;
            }
            if(tmjd == imjd+1){
              x2 = atof(line.substr(18,9).c_str());
              y2 = atof(line.substr(37,9).c_str());
              X2 = atof(line.substr(97,9).c_str())/1000;
              Y2 = atof(line.substr(116,9).c_str())/1000;
              dut2 = atof(line.substr(58,10).c_str());
              smjd = tmjd;
              break;
            }
          }
        }
      myfile.close();
    }

  if(smjd==0 || x2==0){ xp=x1; dX=X1; }
  else{
    xp=(mjd_utc-fmjd)/(smjd-fmjd)*x2+(smjd-mjd_utc)/(smjd-fmjd)*x1;
    dX=(mjd_utc-fmjd)/(smjd-fmjd)*X2+(smjd-mjd_utc)/(smjd-fmjd)*X1;
  }
  
  if(smjd==0 || y2==0){ yp=y1; dY=Y1; }
  else{
    yp=(mjd_utc-fmjd)/(smjd-fmjd)*y2+(smjd-mjd_utc)/(smjd-fmjd)*y1;
    dY=(mjd_utc-fmjd)/(smjd-fmjd)*Y2+(smjd-mjd_utc)/(smjd-fmjd)*Y1;
  }

  xp/=3600*180/DPI; //" to rad, to be compatible to sofa
  yp/=3600*180/DPI; //" to rad, to be compatible to sofa
  dX/=3600*180/DPI; //" to rad, to be compatible to sofa
  dY/=3600*180/DPI; //" to rad, to be compatible to sofa

  if(smjd==0 || dut2==0) return dut1;
  return (mjd_utc-fmjd)/(smjd-fmjd)*dut2+(smjd-mjd_utc)/(smjd-fmjd)*dut1;
}

double CalculateDT(const double year){
  // from http://eclipse.gsfc.nasa.gov/SEhelp/deltatpoly2004.html
  // based on Morrison, L. and Stephenson, F. R., "Historical Values of the Earth's Clock Error DeltaT and the Calculation of Eclipses", J. Hist. Astron., Vol. 35 Part 3, August 2004, No. 120, pp 327-336 (2004).
  if(year<-500){
    double u=(year-1820)/100;
    return -20 + 32*u*u;
      
  }else if(year<500){
    double u=year/100;
    return 10583.6 + u*(-1014.41 + u*(33.78311 + u*(-5.952053 + u*(-0.1798452 + u*(0.022174192 + u*0.0090316521)))));
      
  }else if(year<1600){
    double u=(year-1000)/100;
    return 1574.2 + u*(-556.01 + u*(71.23472 + u*(0.319781 + u*(-0.8503463 + u*(-0.005050998 + u*0.0083572073)))));
      
  }else if(year<1700){
    double t=year-1600;
    return 120 + t*(-0.9808 + t*(-0.01532 + t/7129));
      
  }else if(year<1800){
    double t=year-1700;
    return 8.83 + t*(0.1603 + t*(-0.0059285 + t*(0.00013336 - t/1174000)));
      
  }else if(year<1860){
    double t=year-1800;
    return 13.72 + t*(-0.332447 + t*(0.0068612 + t*(0.0041116 + t*(-0.00037436 + t*(0.0000121272 + t*(-0.0000001699 + t*0.000000000875))))));
      
  }else if(year<1900){
    double t=year-1860;
    return 7.62 + t*(0.5737 + t*(-0.251754 + t*(0.01680668 + t*(-0.0004473624 + t/233174))));
      
  }else if(year<1920){
    double t=year-1900;
    return -2.79 + t*(1.494119 + t*(-0.0598939 + t*(0.0061966 - 0.000197 * t)));
      
  }else if(year<1941){
    double t=year-1920;
    return 21.20 + t*(0.84493 + t*(-0.076100 + 0.0020936 * t));
      
  }else if(year<1961){
    double t=year-1950;
    return 29.07 + t*(0.407 - t/233 + t*t/2547);
      
  }else if(year<1986){
    double t=year-1975;
    return 45.45 + t*(1.067 - t/260 - t*t/718);
      
  }else if(year<2005){
    double t=year-2000;
    return 63.86 + t*(0.3345 + t*(-0.060374 + t*(0.0017275 + t*(0.000651814 + 0.00002373599 * t))));
      
  }else if(year<2050){
    double t=year-2000;
    return 62.92 + t*(0.32217 + 0.005589 * t);
      
  }else if(year<2150){
    double u=(year-1820)/100;
    return -20 + 32*u*u - 0.5628 * (2150 - year);
      
  }else{
    double u=(year-1820)/100;
    return -20 + 32*u*u;
  }
}

double ReadDT(double year){
  double dt1=0, dt2=0, tyear, fyear=0, syear=0;
  std::string line;
  std::ifstream myfile2("corrections/deltat.preds");
  if(myfile2.is_open())
    {
      while(myfile2.good())
        {
          getline(myfile2,line);
          if(line.length()>20){
            tyear = atof(line.substr(13,8).c_str());
            if(year>tyear && year<=tyear+0.25 && dt1==0){
              dt1 = atof(line.substr(22,7).c_str());
              fyear = tyear;
              continue;
            }
            if(year<=tyear+0.25 && dt1!=0){
              dt2 = atof(line.substr(22,7).c_str());
              syear = tyear;
              break;
            }
          }
        }
      myfile2.close();
    }

  if(fyear==0 || dt1==0 || syear==0 || dt2==0){
    return CalculateDT(year);
  }else
    return (year-fyear)/(syear-fyear)*dt2+(syear-year)/(syear-fyear)*dt1;
}

double ReadDT(const double mjd, const double year){
  double dt1=0, dt2=0, tmjd, fmjd=0, smjd=0;
  std::string line;
  std::ifstream myfile2("corrections/deltat.preds");
  if(myfile2.is_open())
    {
      while(myfile2.good())
        {
          getline(myfile2,line);
          if(line.length()>20){
            tmjd = atof(line.substr(0,12).c_str());
            if(tmjd <= mjd){
              dt1 = atof(line.substr(22,7).c_str());
              fmjd = tmjd;
              continue;
            }
            if(tmjd > mjd){
              dt2 = atof(line.substr(22,7).c_str());
              smjd = tmjd;
              break;
            }
          }
        }
      myfile2.close();
    }

  if(fmjd==0 || smjd==0){
    return CalculateDT(year);
  }else
    return (mjd-fmjd)/(smjd-fmjd)*dt2+(smjd-mjd)/(smjd-fmjd)*dt1;
}


void DownloadFiles(){
  printf("Searching for IERS files... \n");
  time_t rawtime;
  struct stat attrib;
  float fileage;
  size_t filesize;
  std::string filename;
  time(&rawtime);

  // //based on McCarthy and Babcock (1986)
  // fileage = std::numeric_limits<float>::infinity();
  // filesize= 0;
  // if(!stat("corrections/historic_deltat.data", &attrib)){
  //   fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
  //   filesize= attrib.st_size;
  //   printf("historic_deltat.data found (%f days old)\n", fileage);
  // }
  // else
  //   printf("historic_deltat.data not found. Downloading...\n");
  // if(filesize==0 || fileage>180){
  //   if(!downloadFile("http://maia.usno.navy.mil/ser7/historic_deltat.data","corrections/historic_deltat.data"))
  //     downloadFile("http://toshi.nofs.navy.mil/ser7/historic_deltat.data","corrections/historic_deltat.data");
  // }

/*fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("corrections/deltat.preds", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("deltat.preds found (%f days old)\n", fileage);
  }
  else
    printf("deltat.preds not found. Downloading...\n");
  if(filesize==0 || fileage>45){
    if(!downloadFile("ftp://cddis.gsfc.nasa.gov/pub/products/iers/deltat.preds","corrections/deltat.preds"))
      if(!downloadFile("http://maia.usno.navy.mil/ser7/deltat.preds","corrections/deltat.preds"))
	downloadFile("http://toshi.nofs.navy.mil/ser7/deltat.preds","corrections/deltat.preds");
  }

  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("corrections/tai-utc.dat", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("tai-utc.dat found (%f days old)\n", fileage);
  }
  else
    printf("tai-utc.dat not found. Downloading...\n");
  if(filesize==0 || fileage>90){
    if(!downloadFile("ftp://cddis.gsfc.nasa.gov/pub/products/iers/tai-utc.dat","corrections/tai-utc.dat"))
      if(!downloadFile("http://maia.usno.navy.mil/ser7/tai-utc.dat","corrections/tai-utc.dat"))
	downloadFile("http://toshi.nofs.navy.mil/ser7/tai-utc.dat","corrections/tai-utc.dat");
  }*/

  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("corrections/finals2000A.data", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("finals2000A.data found (%f days old)\n", fileage);
  }
  else
    printf("finals2000A.data not found. Downloading...\n");
  if(filesize==0 || fileage>7){
    if(!downloadFile("ftp://cddis.gsfc.nasa.gov/pub/products/iers/finals2000A.data","corrections/finals2000A.data",true))
      if(!downloadFile("ftp://ftp.iers.org/products/eop/rapid/standard/finals2000A.data","corrections/finals2000A.data",true))
	if(!downloadFile("http://maia.usno.navy.mil/ser7/finals2000A.data","corrections/finals2000A.data",true))
	  downloadFile("http://toshi.nofs.navy.mil/ser7/finals2000A.data","corrections/finals2000A.data",true);        
  }

  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("corrections/finals2000A.daily", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("finals2000A.daily found (%f days old)\n", fileage);
  }
  else
    printf("finals2000A.daily not found. Downloading...\n");
  if(filesize==0 || fileage>2){
    if(!downloadFile("ftp://cddis.gsfc.nasa.gov/pub/products/iers/finals2000A.daily","corrections/finals2000A.daily",true))
      if(!downloadFile("ftp://ftp.iers.org/products/eop/rapid/daily/finals2000A.daily","corrections/finals2000A.daily",true))
	if(!downloadFile("http://maia.usno.navy.mil/ser7/finals2000A.daily","corrections/finals2000A.daily",true))
	  downloadFile("http://toshi.nofs.navy.mil/ser7/finals2000A.daily","corrections/finals2000A.daily",true);
  }

  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("corrections/eopc04_IAU2000.62-now", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("eopc04_IAU2000.62-now found (%f days old)\n", fileage);
  }
  else
    printf("eopc04_IAU2000.62-now not found. Downloading...\n");
  if(filesize==0 || fileage>7){
    if(!downloadFile("ftp://ftp.iers.org/products/eop/long-term/c04_14/iau2000/eopc04_14_IAU2000.62-now","corrections/eopc04_IAU2000.62-now",true))
      downloadFile("ftp://hpiers.obspm.fr/iers/eop/eopc04/eopc04_IAU2000.62-now","corrections/eopc04_IAU2000.62-now",true);
  }
  
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("corrections/TTBIPM.act", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("TTBIPM.act found (%f days old)\n", fileage);
  }
  else
    printf("TTBIPM.act not found. Downloading...\n");
  if(filesize==0 || fileage>30){
    std::string url = "ftp://ftp2.bipm.org/pub/tai/ttbipm/";
    listFtp(url.data());
    url+=readFtpList("TTBIPM");
    downloadFile(url.data(),"corrections/TTBIPM.act");
  }

  // fileage = std::numeric_limits<float>::infinity();
  // filesize= 0;
  // if(!stat("corrections/ealtai_act.ar", &attrib)){
  //   fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
  //   filesize= attrib.st_size;
  //   printf("ealtai_act.ar found (%f days old)\n", fileage);
  // }
  // else
  //   printf("ealtai_act.ar not found. Downloading...\n");
  // if(filesize==0 || fileage>30){
  //   std::string url = "ftp://ftp2.bipm.org/pub/tai/scale/EALTAI/";
  //   listFtp(url.data());
  //   url+=readFtpList("ealtai");
  //   downloadFile(url.data(),"corrections/ealtai_act.ar");
  // }
}

void DownloadEphemerides(){
  //this is not necessarty for normal use
  printf("Searching for Ephemeris files... \n");
  time_t rawtime;
  struct stat attrib;
  float fileage;
  size_t filesize;
  std::string filename;
  time(&rawtime);

  // download ephemeris files. tested with: de430.bsp, de431.bsp, de432.bsp, de432s.bsp, de436s.bsp, de438s.bsp and de440s.bsp
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/de440s.bsp", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("de440s.bsp found (%f days old)\n", fileage);
  }
  else
    printf("de440s.bsp not found. Downloading...\n");
  if(filesize==0){
    if(!downloadFile("ftp://ssd.jpl.nasa.gov/pub/eph/planets/bsp/de440s.bsp","data/de440s.bsp"))
      downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/de440s.bsp","data/de440s.bsp");
  }

  // download iau orientation file
  // IAU_EARTH, IAU_MOON, IAU_MARS,...
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/pck000xx.tpc", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("pck000xx.tpc found (%f days old)\n", fileage);
  }
  else
    printf("pck000xx.tpc not found. Downloading...\n");
  if(filesize==0  || fileage>360){
    std::string url = "ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/";
    listFtp(url.data());
    url+=readFtpList("pck000","tpc");
    downloadFile(url.data(),"data/pck000xx.tpc");
  }
  return;

  // download high precision earth orientation file
  // ITRF93
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/earth_latest_high_prec.bpc", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("earth_latest_high_prec.bpc found (%f days old)\n", fileage);
  }
  else
    printf("earth_latest_high_prec.bpc not found. Downloading...\n");
  if(filesize==0 || fileage>7){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/earth_latest_high_prec.bpc","data/earth_latest_high_prec.bpc",true);
  }

  // download EARTH_FIXED switch (we have the version we need already included)
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/earth_fixed.tf", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("earth_fixed.tf found (%f days old)\n", fileage);
  }
  else
    printf("earth_fixed.tf not found. Downloading...\n");
  if(filesize==0){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/earth_fixed.tf","data/earth_fixed.tf");
  }

  // download mass parameters (we do not use this file)
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/gm_de431.tpc", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("gm_de431.tpc found (%f days old)\n", fileage);
  }
  else
    printf("gm_de431.tpc not found. Downloading...\n");
  if(filesize==0){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/gm_de431.tpc","data/gm_de431.tpc");
  }
  
  // download leap second file (we do not use this file)
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/naif00xx.tls", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("naif00xx.tls found (%f days old)\n", fileage);
  }
  else
    printf("naif00xx.tls not found. Downloading...\n");
  if(filesize==0  || fileage>180){
    std::string url = "ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/";
    listFtp(url.data());
    url+=readFtpList("naif00","tls");
    downloadFile(url.data(),"data/naif00xx.tls");
  }

  // download mars satellites
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/mar097.bsp", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("mar097.bsp found (%f days old)\n", fileage);
  }
  else
    printf("mar097.bsp not found. Downloading...\n");
  if(filesize==0){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/satellites/mar097.bsp","data/mar097.bsp");
  }
  
  // download jupiters satellites
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/jup310.bsp", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("jup310.bsp found (%f days old)\n", fileage);
  }
  else
    printf("jup310.bsp not found. Downloading...\n");
  if(filesize==0){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/satellites/jup310.bsp","data/jup310.bsp");
  }

  // download saturns satellites
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/sat375.bsp", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("sat375.bsp found (%f days old)\n", fileage);
  }
  else
    printf("sat375.bsp not found. Downloading...\n");
  if(filesize==0){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/satellites/sat375.bsp","data/sat375.bsp");
  }

  // download uranus satellites
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/ura111.bsp", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("ura111.bsp found (%f days old)\n", fileage);
  }
  else
    printf("ura111.bsp not found. Downloading...\n");
  if(filesize==0){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/satellites/ura111.bsp","data/ura111.bsp");
  }

  // download neptunes satellites
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/nep081.bsp", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("nep081.bsp found (%f days old)\n", fileage);
  }
  else
    printf("nep081.bsp not found. Downloading...\n");
  if(filesize==0){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/satellites/nep081.bsp","data/nep081.bsp");
  }

  // download plutos satellites
  fileage = std::numeric_limits<float>::infinity();
  filesize= 0;
  if(!stat("data/plu055.bsp", &attrib)){
    fileage = difftime(rawtime,attrib.st_mtime)/24/3600;
    filesize= attrib.st_size;
    printf("plu055.bsp found (%f days old)\n", fileage);
  }
  else
    printf("plu055.bsp not found. Downloading...\n");
  if(filesize==0){
    downloadFile("ftp://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/satellites/plu055.bsp","data/plu055.bsp");
  }

}
